
#ifndef _CRACKERS_INDEX_H_
#define _CRACKERS_INDEX_H_

#ifdef LIBCRACKERS
/* Type definitions shared within the crackers module/library */

/* used to keep track of the bats that use a given map set within a query
   With this information we can improve the alignment procedure of incremental sideways cracking
   by avoiding alignment operations not necessary for the current query 
   All bats required for the current query are kept in these nodes as a list in the CrackerIndexNode of the cracker bat
  of the head of the used map set */

struct referencesNodeFM{
	int 		batId;
	int 		tailBatId;
	struct referencesNodeFM *next;
};

struct referencesNode{
	int 		batId;
	oid 		cursor;
	struct referencesNode *next;
};

typedef struct referencesNode refNode; 

struct Node{
	/* Tree related */
        struct Node  	*left;
        struct Node  	*right;
        int	      	height;
	bit	      	head;

	/* Information stored at this node */
        oid	      	position;		/* The position in the BAT of the index where the bounds are stored */
	bit 	      	inclusive;	  
	bit	      	deleted;		/* If true then this node is deleted and can be reused */
        struct Node  	*previous;
	bit 	      	isPreviousSmaller;	
	oid	      	hols; 		  	/* # of hols before this piece */

	/* Joins */	
	int 	      	slice;		  	/* A view between the position of this index entry and the next one
						   Keeping views around allows us to reuse hash tables built for a view for joins */	

	/* Info needed for incremental maps */
	int 		pieceBAT;		/* The BAT id of the  map for this piece */
	bit		localBound;		/* TRUE if this bound exists only in this piece. FALSE if it exists in the cracker BAT too */
	bit 		first;			/* TRUE if this is the first bound in this piece. FALSE if it is the last. 
							Valid only if localBound=FALSE */
	oid 		indexLocal;		/* Local crack positions */

	 /* Used in the storage management procedure*/ 
	struct Node     *nextPiece;		/*  It identifies the next piece, i.e., the next first node in the map*/
	lng 		creationTimestamp;
	lng 		accesses;
	lng		dropped;
	oid 		tuples;

	/* Used on the cracker BAT side in the case of incremental maps */
	struct referencesNode *references;	/* maps that use this piece */
	int 		tape;			/* Tape holding previous cracks, (value,inclusive) */
};

typedef struct {
			/* Core cracking */
	int 		bid;   		/* The base BAT */
	int 		cbid;   	/* The cracker column/BAT i.e., the copy on which we actually crack */

	int 		cid;   		/* The cracker index for BAT */
	struct Node 	*Tree; 		/* The AVL-tree part of the cracker index */

			/* Updates */
	int 		iid;   		/* The pending insertions column */
	int 		did;   		/* The pending deletions column */

	bit		reCreate;	/* Indicates whether we need to recreate the index if we chose to forget it */

	sht		mergeInsertions;/* Indicates wether there are insertions to merge -->  -1 no insertions,
												0 complete merge,
												1 gradually,
												2 ripple */
	sht		mergeDeletions; /* Indicates wether there are insertions to merge -->  -1 no insertions,
												0 complete merge,
												1 gradually,
												2 ripple */
	bit 		deleteNodes;    /* If set true, merging operations will delete nodes 
					   form the cracker index if this makes things easier */

			/* Sideways cracking */
	int 		pbid; 		/* The cracker map. Together with the bid, they uniquely identify 
				           a cracker map when pid is -1 then chid stores the cracker 
                                           tape for all the cracker maps where bid is head 
				           hiid is a [bit,bit] BAT that stores inclusive information 
					   for each bound stored in chid */

	oid 		nextCrack;	/* Points to the next cracking operation that should be applied 
                                           in this map to sync it */
	int 		chid;
	int 		hiid;

	struct referencesNode *references;
	lng 		creationTimestamp;
	lng 		accesses;
	lng		dropped;
	oid 		tuples;

} CrackerIndexNode;

/* Signatures shared within the crackers module/library */

int existsCrackerIndex(int bid);





/* CreateNewIndex */
int newCrackerIndex_chr(int bid, int cbid);
void reCreateMap_chr(int position);

/* InsertIndexElements */
void addCrackerIndex_chr(int m, chr *value, bit inclusive, oid index, BAT * b);
struct Node *  addCrackerIndexIncrementalSideways_chr(int m, chr value, bit inclusive, oid indexLocal, oid index, BAT * b, bit localBound, bit first, int pieceBAT);




/* CreateNewIndex */
int newCrackerIndex_sht(int bid, int cbid);
void reCreateMap_sht(int position);

/* InsertIndexElements */
void addCrackerIndex_sht(int m, sht *value, bit inclusive, oid index, BAT * b);
struct Node *  addCrackerIndexIncrementalSideways_sht(int m, sht value, bit inclusive, oid indexLocal, oid index, BAT * b, bit localBound, bit first, int pieceBAT);




/* CreateNewIndex */
int newCrackerIndex_int(int bid, int cbid);
void reCreateMap_int(int position);

/* InsertIndexElements */
void addCrackerIndex_int(int m, int *value, bit inclusive, oid index, BAT * b);
struct Node *  addCrackerIndexIncrementalSideways_int(int m, int value, bit inclusive, oid indexLocal, oid index, BAT * b, bit localBound, bit first, int pieceBAT);




/* CreateNewIndex */
int newCrackerIndex_lng(int bid, int cbid);
void reCreateMap_lng(int position);

/* InsertIndexElements */
void addCrackerIndex_lng(int m, lng *value, bit inclusive, oid index, BAT * b);
struct Node *  addCrackerIndexIncrementalSideways_lng(int m, lng value, bit inclusive, oid indexLocal, oid index, BAT * b, bit localBound, bit first, int pieceBAT);




/* CreateNewIndex */
int newCrackerIndex_flt(int bid, int cbid);
void reCreateMap_flt(int position);

/* InsertIndexElements */
void addCrackerIndex_flt(int m, flt *value, bit inclusive, oid index, BAT * b);
struct Node *  addCrackerIndexIncrementalSideways_flt(int m, flt value, bit inclusive, oid indexLocal, oid index, BAT * b, bit localBound, bit first, int pieceBAT);




/* CreateNewIndex */
int newCrackerIndex_dbl(int bid, int cbid);
void reCreateMap_dbl(int position);

/* InsertIndexElements */
void addCrackerIndex_dbl(int m, dbl *value, bit inclusive, oid index, BAT * b);
struct Node *  addCrackerIndexIncrementalSideways_dbl(int m, dbl value, bit inclusive, oid indexLocal, oid index, BAT * b, bit localBound, bit first, int pieceBAT);




/* CreateNewIndex */
int newCrackerIndex_date(int bid, int cbid);
void reCreateMap_date(int position);

/* InsertIndexElements */
void addCrackerIndex_date(int m, date *value, bit inclusive, oid index, BAT * b);
struct Node *  addCrackerIndexIncrementalSideways_date(int m, date value, bit inclusive, oid indexLocal, oid index, BAT * b, bit localBound, bit first, int pieceBAT);







#endif

/* Exported signatures */

crackers_export str CRKprintCrackerIndexBATpart(int *k, int *bid);
crackers_export str CRKprintAVLTree_int(int *k, int *bid);
crackers_export str CRKprintCrackerBAT(int *k, int *bid);
crackers_export str CRKgetCrackerBAT(int *vid, int *bid);




#endif /* _CRACKERS_INDEX_H */

